#ifndef __precond_cxx
#define __precond_cxx


#include "precond.h"
#include <vector>
#include <limits>
#include "boost/numeric/ublas/io.hpp"


extern "C" { 
//#include "clapack.h"
int dgesv_(int *n, int *nrhs, double *a, int *lda,
	   int *ipiv, double *b, int *ldb, int *info);
}

#include "ANN/ANN.h"



namespace fmi
{
   

   Preconditioner::VectorType
   Preconditioner::
   solve(const VectorType& v) const
   {
      if(!m_PreconditioningFlag)
	return v;
      
      // Multiply v by the preconditioner
      
      int size = m_NumRowsInOperator;

      ScalarType sum;
      
      VectorType r(size);
      
      int k = 0;
      for( ;k<m_NumberOfPoints;k++)
      {
	 sum = 0.0;
	 
	 int i=0;
	 for( ;i<m_NumberOfCardinalPoints;i++)
         sum += m_CardinalPointCoefs(k,i) * v[ m_CardinalPointIds(k,i) ];

	 for( ;i<m_NumberOfCoefficients;i++)
	 {
         sum += m_CardinalPointCoefs(k,i)*v[m_NumberOfPoints + 
						(i-m_NumberOfCardinalPoints) ];

	    //PRINT(m_NumberOfPoints + (i-m_NumberOfCardinalPoints));
	    //PRINT(m_CardinalPointCoefs(k,i));
	 }

	 r[k] = sum;
      }
      for( ;k<size;k++)
      {
	 sum = 0.0;
	 
	 for(int i=0;i<m_NumberOfPoints;i++)
	 {
         sum += m_CardinalPointCoefs(i,k-m_NumberOfPoints+
                                     m_NumberOfCardinalPoints) * v[i];
	 }

	 //PRINT(k-m_NumberOfPoints+m_NumberOfCardinalPoints);
	 //PRINT(k);

	 //PRINT(sum);
	 r[k] = sum;
	 //r[k] = 0.0;
	 //r[k] = v[k];
      }
      
      //PRINT(size);
      //PRINT(m_NumberOfPoints);

      
      return r;
	
   }



   // For GMRES iteration to work, the matrix must be 
   // well conditioned. Unfortunately in its current form it is not. 
   // To make it well conditioned we need to transform to a basis of 
   // approximate cardinal splines as described in the paper by
   // Beatson. 
   // To do this we need to solve a small linear system for each of the N 
   // points. Basically we assume that each cardinal spline can be expressed
   // as a linear combination of only b of the original (multiquadric) 
   // basis functions where b << N. The b basis functions that we choose
   // are those associated with the closest b points to the current point.
   // 
   // Thus now we form an bxB matrix were each row of the matrix
   // contains the id's of the closest b points to each of the N points
   // 
   // Note: Associated with each of the landmarks is a list
   // closest points (and the special points). Each list is stored
   // in a seperate row of the matrix m_CardinalPointIds
   // In addition each point has associated with it a set of
   // coefficients which are the coefficients of the terms in the sum
   // constituting the approximate cardinal function
      
   // Step 1. compute the cardinal point ids for each source landmark
   int
   Preconditioner::
   SetupPreconditioner(const OperatorType& A, const int num_rows,
		       const ScalarType* landmarks,const int numberOfPoints)
   {

      m_NumberOfPoints = numberOfPoints;
      
      m_PreconditioningFlag = true;

      m_NumberOfClosestPoints = 123;
      //m_NumberOfClosestPoints = 200;

      m_NumberOfSpecialPoints = 27;
      

      m_NumRowsInOperator = num_rows;


      if(m_NumberOfPoints <= m_NumberOfClosestPoints + m_NumberOfSpecialPoints)
      {
	 std::cerr << "Warning: Too few points for preconditioning. " 
		   << "Continuing without preconditioning." << std::endl;
	 m_PreconditioningFlag = false;
	 return -1;
      }


      //m_P.resize( A.size1() , A.size2() );
      //m_P *= 0.0; // initialize to zero

      // Note: The last 27 columns of the m_CardinalPointIds matrix 
      //(for each row) will contain the point ids of the 27 special points.

      m_NumberOfCardinalPoints = 
	 m_NumberOfClosestPoints + m_NumberOfSpecialPoints;
      m_CardinalPointIds.resize(m_NumberOfPoints, m_NumberOfCardinalPoints);


      // The following number is the total number of coefficients
      // of each cardinal basis function. It is equal to the number of
      // cardinal points plus the number of coefficients of the 
      // extra polynomial (which depends on its degree)

      m_NumberOfCoefficients = 
	 m_NumberOfCardinalPoints + (num_rows - m_NumberOfPoints);
      m_CardinalPointCoefs.resize(m_NumberOfPoints,m_NumberOfCoefficients);
      //PRINT(m_NumberOfCoefficients);

      // Now run the ANN program which computes the b closest points
      //and the special points for each of the source points
      this->ComputeCardinalPointIds(landmarks);
      //PRINT(m_CardinalPointIds);

      // Now iterate over all the source points and compute the 
      // new basis functions in terms of the old by solving a 
      // small linear system. Once we have the coefficients, 
      // add them to the big matrix m_P
   
      MatrixType V; // The BxB matrix
      //VectorType Y; // the solution 
      VectorType C; // right side

      V.resize(m_NumberOfCoefficients,m_NumberOfCoefficients);
      //Y.resize(m_NumberOfCoefficients);
      C.resize(m_NumberOfCoefficients);

   

      for(int k=0;k<m_NumberOfPoints;k++)
      {

	 // Setup the matrix V
	 
	 for(int i=0;i<m_NumberOfCardinalPoints;i++)
	 {
	    int ptId = m_CardinalPointIds(k,i);
	    
	    int j = i;
	    for(;j<m_NumberOfCardinalPoints;j++)
	    {
	 
	       int ptId2 = m_CardinalPointIds(k,j);

	       V(j,i) = V(i,j) = A(ptId,ptId2);

	       //if(ptId == ptId2)
	       //  std::cout << 1;
	       //assert(A(ptId,ptId2) == A(ptId2,ptId));
	       //DISPLAY(A(ptId,ptId2));
	       //if(i!=j && V(i,j)== 0.0) 
		 //{
		   // TPoint p = landmarks[i];
		   // TPoint p2 = landmarks[j];
		    //PRINT(p[0] << p[1] << p[2]);
		    //PRINT(p2[0] << p2[1] << p2[2]);
		    //PRINT(k);
		    //PRINT(i);
		    //PRINT(j);
		    //PRINT(ptId);
		    //PRINT(ptId2);
		    //PRINT(m_CardinalPointIds(k,i));
		    //PRINT(m_CardinalPointIds(k,j));
	       
		   // assert(V(i,j)!= 0.0);
		// }
	       
	    }
	    

	    
	    for(int c=m_NumberOfPoints;c<num_rows;c++)
	    {
	       V(j,i) = V(i,j) = A(ptId,c);
	       j++;
	    }
	    

	    C[i] = 0.0;


	 } 
	 
	 
	 // Fill the bottom right part of the matrix V with zeros
	 for(int i=m_NumberOfCardinalPoints;i<m_NumberOfCoefficients;i++)
	 {
	   for(int j=m_NumberOfCardinalPoints;j<m_NumberOfCoefficients;j++)
	     V(i,j) = 0.0;
	   C[i] = 0.0;
	 }

	 // Setup the vector C
	 //for(int i=0;i<m_NumberOfCoefficients;i++) 
	 //  C[i] = 0.0;
	 C[0] = 1.0;
	 

	 //for(int i=0;i<m_NumberOfCoefficients;i++)
	 //   for(int j=0;j<m_NumberOfCoefficients;j++)
	 //     assert(V(i,j)== V(j,i));

	 //PRINT(k);
	 this->SolveLinearSystem(V,C);

	 // Store these coefficients in the next row of the
	 // matrix m_CardinalPointCoefs

	 for(int i=0;i<m_NumberOfCoefficients;i++)
         m_CardinalPointCoefs(k,i) = C[i];


      } // end of outermost for loop

      return 0;
   }
   

   
   
   
   
   // Note the matrix A gets overwritten with the LU factorization and
   // vector b gets overwritten with the solution
   void
   Preconditioner::
   SolveLinearSystem(MatrixType& A,VectorType& B) const
   {
   
      // Solve the system using a lapack routine

      //PRINT(A);
      //PRINT(B);

      int N = B.size();
      int IPIV[N];

      assert(sizeof(B[0]) == 8);
      int INFO;

      //INFO = clapack_dgesv(CblasRowMajor, N, 1, &A[0][0], N, IPIV, &B[0], N);

      int NRHS = 1;

      int LDA = N;
      int LDB = N;

      // Since the matrix A is symmetric we don't need to worry about
      // row/column order stuff
      dgesv_(&N, &NRHS, &A(0,0), &LDA, &IPIV[0], &B[0], &LDB, &INFO);


      if (INFO != 0)
      {
	 PRINT(A);
	 PRINT(B);
	 std::cout << "failure with error " << INFO << std::endl;
	 assert(0);
	 abort();
      }




   }
   


   
   
   void
   Preconditioner::
   ComputeCardinalPointIds(const ScalarType* landmarks)
   {

      int i,j;

      const int m_NumberOfCardinalPoints = 
	m_NumberOfClosestPoints + m_NumberOfSpecialPoints;

      // Now run the ANN program which computes the b closest points
      // for each of the source points

      int                 n_pts;               // actual number of data points
      ANNpointArray       data_pts;            // data points
      ANNpoint            query_pt;            // query point
      ANNidxArray         nn_idx;              // near neighbor indices
      ANNdistArray        dists;               // near neighbor distances
      ANNkd_tree          *the_tree;           // search structure
      double              eps     = 0;         // error bound
    
      query_pt = annAllocPt(3);                 // allocate query point
      data_pts = annAllocPts(m_NumberOfPoints, 3); // allocate data points
      nn_idx = new ANNidx[m_NumberOfCardinalPoints+1];// allocate near neigh 
                                                      //indices
      dists = new ANNdist[m_NumberOfCardinalPoints+1];// allocate near 
                                                      //neighbor dists


      const ScalarType* sourcepoint;            // read data points
      for(n_pts=0;n_pts<m_NumberOfPoints;n_pts++)
      {

	 sourcepoint = &landmarks[3*n_pts];

	 data_pts[n_pts][0] = sourcepoint[0];
	 data_pts[n_pts][1] = sourcepoint[1];
	 data_pts[n_pts][2] = sourcepoint[2];
            
      } 

    
      the_tree = new ANNkd_tree(     // build search structure
	 data_pts,                   // the data points
	 m_NumberOfPoints,           // number of points
	 3);                         // dimension of space


      // First determine the 27 special points by using ANN
      // to find the closest points to the 27 points of a 
      // 3x3x3 grid over the data rectangle

      // First we need to compute the points on the 3x3x3 grid
      // We therefore need to compute a bounding box enclosing all
      // of the points. Denote this bounding box by extent[6]

      // The point Id's of the special points
      int specialPointsIds[m_NumberOfSpecialPoints];
  

      ScalarType bounding_box[6];
      this->ComputeBoundingBox(landmarks,bounding_box);

      ScalarType xmin = bounding_box[0];
      ScalarType xmax = bounding_box[1];
      ScalarType ymin = bounding_box[2];
      ScalarType ymax = bounding_box[3];
      ScalarType zmin = bounding_box[4];
      ScalarType zmax = bounding_box[5];

      // The middle point in each dimension
      ScalarType xmid = (xmin + xmax)/2.0;
      ScalarType ymid = (ymin + ymax)/2.0;
      ScalarType zmid = (zmin + zmax)/2.0;

      //ScalarType grid[m_NumberOfSpecialPoints][3] = {
	// {xmin,ymin,zmin},{xmid,ymin,zmin},{xmax,ymin,zmin},
	// {xmin,ymid,zmin},{xmid,ymid,zmin},{xmax,ymid,zmin},
	// {xmin,ymax,zmin},{xmid,ymax,zmin},{xmax,ymax,zmin},
	// {xmin,ymin,zmid},{xmid,ymin,zmid},{xmax,ymin,zmid},
	// {xmin,ymid,zmid},{xmid,ymid,zmid},{xmax,ymid,zmid},
	// {xmin,ymax,zmid},{xmid,ymax,zmid},{xmax,ymax,zmid},
	// {xmin,ymin,zmax},{xmid,ymin,zmax},{xmax,ymin,zmax},
	// {xmin,ymid,zmax},{xmid,ymid,zmax},{xmax,ymid,zmax},
	// {xmin,ymax,zmax},{xmid,ymax,zmax},{xmax,ymax,zmax} };
      MatrixType grid(m_NumberOfSpecialPoints,3);
	//{xmin,ymin,zmin},
	//{xmid,ymin,zmin},
	//{xmax,ymin,zmin},
      grid(0,0)=xmin; grid(0,1)=ymin; grid(0,2)=zmin;
      grid(1,0)=xmid; grid(1,1)=ymin; grid(1,2)=zmin;
      grid(2,0)=xmax; grid(2,1)=ymin; grid(2,2)=zmin;
	//{xmin,ymid,zmin},
	//{xmid,ymid,zmin},
	//{xmax,ymid,zmin},
      grid(3,0)=xmin; grid(3,1)=ymid; grid(3,2)=zmin;
      grid(4,0)=xmid; grid(4,1)=ymid; grid(4,2)=zmin;
      grid(5,0)=xmax; grid(5,1)=ymid; grid(5,2)=zmin;
	//{xmin,ymax,zmin},
	//{xmid,ymax,zmin},
	//{xmax,ymax,zmin},
      grid(6,0)=xmin; grid(6,1)=ymax; grid(6,2)=zmin;
      grid(7,0)=xmid; grid(7,1)=ymax; grid(7,2)=zmin;
      grid(8,0)=xmax; grid(8,1)=ymax; grid(8,2)=zmin;
	//{xmin,ymin,zmid},
	//{xmid,ymin,zmid},
	//{xmax,ymin,zmid},
      grid(9,0) =xmin; grid(9,1) =ymin; grid(9,2) =zmid;
      grid(10,0)=xmid; grid(10,1)=ymin; grid(10,2)=zmid;
      grid(11,0)=xmax; grid(11,1)=ymin; grid(11,2)=zmid;
	//{xmin,ymid,zmid},
	//{xmid,ymid,zmid},
	//{xmax,ymid,zmid},
      grid(12,0)=xmin; grid(12,1)=ymid; grid(12,2)=zmid;
      grid(13,0)=xmid; grid(13,1)=ymid; grid(13,2)=zmid;
      grid(14,0)=xmax; grid(14,1)=ymid; grid(14,2)=zmid;
	//{xmin,ymax,zmid},
	//{xmid,ymax,zmid},
	//{xmax,ymax,zmid},
      grid(15,0)=xmin; grid(15,1)=ymax; grid(15,2)=zmid;
      grid(16,0)=xmid; grid(16,1)=ymax; grid(16,2)=zmid;
      grid(17,0)=xmax; grid(17,1)=ymax; grid(17,2)=zmid;
	//{xmin,ymin,zmax},
	//{xmid,ymin,zmax},
	//{xmax,ymin,zmax},
      grid(18,0)=xmin; grid(18,1)=ymin; grid(18,2)=zmax;
      grid(19,0)=xmid; grid(19,1)=ymin; grid(19,2)=zmax;
      grid(20,0)=xmax; grid(20,1)=ymin; grid(20,2)=zmax;
	//{xmin,ymid,zmax},
	//{xmid,ymid,zmax},
	//{xmax,ymid,zmax},
      grid(21,0)=xmin; grid(21,1)=ymid; grid(21,2)=zmax;
      grid(22,0)=xmid; grid(22,1)=ymid; grid(22,2)=zmax;
      grid(23,0)=xmax; grid(23,1)=ymid; grid(23,2)=zmax;
	//{xmin,ymax,zmax},
	//{xmid,ymax,zmax},
	//{xmax,ymax,zmax}
      grid(24,0)=xmin; grid(24,1)=ymax; grid(24,2)=zmax;
      grid(25,0)=xmid; grid(25,1)=ymax; grid(25,2)=zmax;
      grid(26,0)=xmax; grid(26,1)=ymax; grid(26,2)=zmax;


      for(i=0;i<m_NumberOfSpecialPoints;i++)
      {
	 query_pt[0] = grid(i,0);
	 query_pt[1] = grid(i,1);
	 query_pt[2] = grid(i,2);

	 bool foundUniquePoint = false;

	 // Number of points to search for. If we can't find a unique point 
	 // on the first try then increase the number of points to search for
	 // by one for the next try.
	 n_pts = 1; 

	 while(foundUniquePoint == false)
	 {

	    the_tree->annkSearch(              // search
	       query_pt,                       // query point
	       n_pts,                          // number of near neighbors
	       nn_idx,                         // nearest neighbors (returned)
	       dists,                          // distance (returned)
	       eps);                           // error bound

	    // if the new special point is already on our list
	    // then we need to search for another point
	    if( std::find(specialPointsIds,
			  specialPointsIds+i,
			  nn_idx[n_pts-1])    ==  specialPointsIds+i )
	    {
	       specialPointsIds[i] = nn_idx[n_pts-1];
	       foundUniquePoint = true;
	       //PRINT(specialPointsIds[i]);
	    }
	    else
	    {
	       n_pts++;
	    }

	 }

	 //sourcepoint = landmarks->GetPoints()->GetElement(nn_idx[0]);
	 //std::cout << grid(i,0) << " " << grid(i,1) << " "<< grid(i,2)<<"->";
	 //std::cout << sourcepoint << std::endl;
	 //PRINT(specialPointsIds[i]);
      }


      for(n_pts=0;n_pts<m_NumberOfPoints;n_pts++)
      {

	 sourcepoint = &landmarks[3*n_pts];
	 query_pt[0] = sourcepoint[0];
	 query_pt[1] = sourcepoint[1];
	 query_pt[2] = sourcepoint[2];
	 //PRINT(sourcepoint[0] << sourcepoint[1] << sourcepoint[2]);
	 //PRINT(n_pts);
	 //PRINT(m_NumberOfPoints);
	 //PRINT(m_NumberOfCardinalPoints);
	 //PRINT(m_NumberOfClosestPoints);

	 // Note: we search for m_NumberOfCardinalPoints points since 
	 // if we were to search for m_NumberOfClosestPoints points
	 // then some of the nearest neighbors found will be the special
	 // points which are already included in our list of cardinal points
	 the_tree->annkSearch(              // search
	    query_pt,                       // query point
	    m_NumberOfCardinalPoints+1,       // number of near neighbors
	    nn_idx,                         // nearest neighbors (returned)
	    dists,                          // distance (returned)
	    eps);                           // error bound

      
	 // Store the closest points found in the next row of 
	 // m_CardinalPointIds excluding those that are among the 27 special 
	 // points since the last 27 columns of each row will contain these 
	 // 27 points

	 // The first point in each line is always the current query point
	 // even if the current query point is one of the special points
	 m_CardinalPointIds(n_pts,0) = nn_idx[0];

	 i=j=1;
	 while(i<m_NumberOfClosestPoints)
	 {
	    // if one of the found closest points is one of the special points
	    // then ignore it
	    if( std::find(specialPointsIds,
			  specialPointsIds+m_NumberOfSpecialPoints,
			  nn_idx[j]) == 
		specialPointsIds+m_NumberOfSpecialPoints )
	    {
           m_CardinalPointIds(n_pts,i) = nn_idx[j];
	       i++;
	       //sourcepoint = landmarks->GetPoints()->GetElement(nn_idx[j]);
	       //std::cout << sourcepoint << std::endl;
	       //DISPLAY(nn_idx[j]);
	    }
	    j++;
	 }
	 //PRINT(" ");

	 // Now add the special points to the last 27 columns of 
	 // each row of m_CardinalPointIds
	 for( ;i<m_NumberOfCardinalPoints;i++)
	 {
         m_CardinalPointIds(n_pts,i) = 
	      specialPointsIds[i-m_NumberOfClosestPoints];
	 }

      
	 // if the current query point is one of the special points then
	 // we will end up having a duplicate value in our list of 
	 // cardinal point ids. Therefore we need to replace that value
	 // with a new closest point id.

	 
	 int *location = 
	   std::find(specialPointsIds,
		     specialPointsIds+m_NumberOfSpecialPoints,
		     n_pts);

	 if( location != specialPointsIds+m_NumberOfSpecialPoints )
	 {
	    while(i<m_NumberOfCardinalPoints+1)
	    {
	       // if the new closest point is one of the special points
	       // then ignore it
	       if( std::find(specialPointsIds,
			     specialPointsIds+m_NumberOfSpecialPoints,
			     nn_idx[j]) == 
		   specialPointsIds+m_NumberOfSpecialPoints )
	       {
               m_CardinalPointIds(n_pts,m_NumberOfClosestPoints+
                                  (location-specialPointsIds)) = 
		     nn_idx[j];
		  i++;
		  //std::cout << sourcepoint << std::endl;
		  //DISPLAY(nn_idx[j]);
	       }
	       j++;
	    }
	    //PRINT(" ");
	 }
	 
	 
	 //assert(0);
	 // fix this
	 // Now sort the current row of cardinal point ids in increasing order
	 // (except for the first point of course)
	 //std::sort(&m_CardinalPointIds(n_pts,1),
	//	   &m_CardinalPointIds(n_pts,1)+m_NumberOfCardinalPoints-1);

      }
   
      //delete stuff here
      annDeallocPt(query_pt);          
      annDeallocPts(data_pts);
      delete [] nn_idx;          
      delete [] dists;            
      delete the_tree;

   }



   // Compute a bounding box of all the points
   void
   Preconditioner::
   ComputeBoundingBox(const ScalarType* landmarks,
		      ScalarType* box) const
   {

      box[0] = box[2] = box[4] = std::numeric_limits<ScalarType>::max();
      box[1] = box[3] = box[5] = std::numeric_limits<ScalarType>::min();
      
      ScalarType p[3];

      //iterate over all the points to find 
      // a bounding box of the points
      for(int i=0;i<m_NumberOfPoints;i++)
      {
	 p[0] = landmarks[i*3];
	 p[1] = landmarks[i*3+1];
	 p[2] = landmarks[i*3+2];

	 if(p[0] < box[0])
	   box[0] = p[0];
	 if(p[0] > box[1])
	   box[1] = p[0];

	 if(p[1] < box[2])
	   box[2] = p[1];
	 if(p[1] > box[3])
	   box[3] = p[1];

	 if(p[2] < box[4])
	   box[4] = p[2];
	 if(p[2] > box[5])
	   box[5] = p[2];

      }
      

      //PRINT(box[0]);
      //PRINT(box[1]);
      //PRINT(box[2]);
      //PRINT(box[3]);
      //PRINT(box[4]);
      //PRINT(box[5]);
      
   }

	
}

#endif
