#ifndef __precond_h
#define __precond_h



#include "boost/numeric/ublas/matrix.hpp"
#include "boost/numeric/ublas/vector.hpp"

#include "rbfmatrix.h"

namespace fmi
{
   

   // This class represents a preconditioner and adds
   // the function solve which is needed when using the preconditioner
   // in the function GMRES(..)

   // To use this class you need to call only one setup function:
   // SetupPreconditioner(...) Then you can call one of the versions of solve.
   class Preconditioner
   {

      public:
	 
	 typedef double ScalarType;
	 typedef boost::numeric::ublas::matrix< ScalarType > MatrixType;
	 typedef boost::numeric::ublas::vector< ScalarType > VectorType;
	 typedef rbfmatrix OperatorType;
	

	 // Takes a reference to the matrix that needs preconditioning

	 // Note that the number of points can be less than num_rows
	 // since there might be extra terms for the affine polynomial
	 // Returns 0 on success and -1 on failure
	 int SetupPreconditioner(const OperatorType& A, const int num_rows,
				 const ScalarType* landmarks,
				 const int numberOfPoints);

	 // multiplies the input vector by the preconditioner
	 // using left preconditioning
	 VectorType solve(const VectorType& v) const;


	 // Constructor
	 Preconditioner(): m_PreconditioningFlag(false) {}
	

      private:

	

	 void ComputeCardinalPointIds(const ScalarType* landmarks);

	 void ComputeBoundingBox(const ScalarType* landmarks,
				 ScalarType* box) const;

	 void SolveLinearSystem(MatrixType& M,VectorType& b) const;
	
	 bool m_PreconditioningFlag;

	 int m_NumberOfPoints;
	 int m_NumberOfClosestPoints;
	 int m_NumberOfSpecialPoints;
	 int m_NumberOfCardinalPoints;
	 int m_NumberOfCoefficients;

	 int m_NumRowsInOperator;

	 typedef boost::numeric::ublas::matrix< int > MatrixIntType;
	 MatrixIntType       m_CardinalPointIds;

	 MatrixType          m_CardinalPointCoefs;

	
   };
	
}

#endif
